from json import dumps
from requests import post
from base64 import b64encode
from math import floor
from ._utils import clamp, get_color_rgb, tuple_to_hex
from ._constants import DisplayType, DIAL_DEFAULT_ITEM
from datetime import datetime
from time import time, sleep
from PIL import Image, ImageTk
import tkinter as tk

class Renderer:
	_size = None
	_max_frames = None
	_debug = False
	_address = None
	_pizoo = None
	_max_frame_speed = 10000  # in ms
	_min_frame_speed = 10     # in ms

	def __init__(self, address, pizzoo, debug):
		'''
		A renderer is an object that is used to render the frames on the device. It can be a real device, an emulator or a static image.
		'''
		self._debug = debug
		self._address = address
		self._pizzoo = pizzoo

	def get_size(self):
		'''
		Returns the size of the device screen.

		Returns:
			int: The size of the device screen.
		'''
		return self._size
	
	def get_max_frames(self):
		'''
		Returns the maximum amount of frames the device buffer can store.

		Returns:
			int: The maximum amount of frames the device can store.	
		'''
		return self._max_frames
	
	def get_max_frame_speed(self):
		'''
		Returns the maximum frame speed of the device in milliseconds.

		Returns:
			int: The maximum frame speed of the device in milliseconds.
		'''
		return self._max_frame_speed
	
	def get_min_frame_speed(self):
		'''
		Returns the minimum frame speed of the device in milliseconds.

		Returns:
			int: The minimum frame speed of the device in milliseconds.
		'''
		return self._min_frame_speed
	
	def get_settings(self):
		'''
		Gets the current settings of the device.
		'''
		raise NotImplementedError
	
	def set_brightness(self, brightness):
		'''
		Sets the brightness of the device.

		Args:
			brightness (int): The brightness level to set. Between 0 and 100.
		'''
		raise NotImplementedError
	
	def switch(self, on=True):
		'''
		Turns the device on or off.

		Args:
			on (bool): Whether to turn the device on or off.
		'''
		raise NotImplementedError

	def render(self, buffer, frame_speed):
		'''
		Renders the buffer on the device.

		Args:
			buffer (list): A list of frames to render on the device.
			frame_speed (int): The speed at which the frames should be displayed.
		'''
		raise NotImplementedError
	
	def compile_node(self, node, parent, inherited_props, node_props):
		'''
		Compiles an XML node into a command that can be executed on the device.

		Args:
			node (Element): The XML node to compile.
			parent (Element): The parent XML node.
			inherited_props (dict): The properties inherited from the parent node.
			node_props (dict): The properties of the current node.
		'''
		return None

	def compile_node_root_options(self, options):
		'''
		Compiles the root options of the XML into a list of commands that can be executed on the device.

		Args:
			options (dict): The options of the root node.
		'''
		return []

	def render_template_items(self, items, use_cache=True):
		'''
		Renders or process a list of items on the device.

		Args:
			items (list): A list of items to render on the device.
			use_cache (bool): Whether to use the cache or not.
		'''
		return None

class Pixoo64Renderer(Renderer):
	__start_countdown_time = -1
	__pic_id = None
	__url = None
	__id_limit = None
	def __init__(self, address, pizzoo, debug):
		'''
		This renderer is used to render the frames on the Divoom Pixoo64 device. It uses the Divoom API to send the frames to the device.
		Also includes some built-in methods for controlling the device like the buzzer, scoreboard, countdown, etc.
		'''
		super().__init__(address, pizzoo, debug)
		self._size = 64
		self._max_frames = 60
		self.__id_limit = 100
		self.__url = f'http://{address}/post'
		self.__pic_id = self.__request('Draw/GetHttpGifId')['PicId']
		if self.__pic_id > self.__id_limit:
			self.__reset_pic_id()

	def __request(self, endpoint, data=None):
		data = {'Command': endpoint, **(data if data else {})}
		result = post(self.__url, dumps(data)).json()
		if result['error_code'] != 0:
			raise Exception(f'Error on request {endpoint} with code \"{result["error_code"]}\"')
		return result
	
	def __reset_pic_id(self):
		try:
			self.__request('Draw/ResetHttpGifId')
			self.__pic_id = 1
		except Exception as e:
			if self._debug: print(e)

	def __send_frame(self, frame_data, speed=1000, frame_number=1, offset=0):
		return self.__request('Draw/SendHttpGif', {
			'PicNum': frame_number,
			'PicWidth': self._size,
			'PicOffset': offset,
			'PicID': self.__pic_id,
			'PicSpeed': speed,
			'PicData': b64encode(bytearray(frame_data)).decode()
		})
	
	def buzzer(self, active=0.5, inactive=0.5, duration=1):
		'''
		Plays a sound on the device buzzer.

		Args:
			active (float): The time in seconds the buzzer is active.
			inactive (float): The time in seconds the buzzer is inactive.
			duration (float): The total time in seconds the buzzer will play.

		Returns:
			None
		'''
		self.__request('Device/PlayBuzzer', {
			'ActiveTimeInCycle': active * 1000,
			'OffTimeInCycle': inactive * 1000,
			'PlayTotalTime': duration * 1000
		})
	
	def set_scoreboard(self, blue_score=0, red_score=0):
		'''
		Sets the scoreboard on the device for every team.

		Args:
			blue_score (int): The score for the blue team.
			red_score (int): The score for the red team.

		Returns:
			None
		'''
		self.__request('Tools/SetScoreBoard', {
			'BlueScore': blue_score,
			'RedScore': red_score
		})

	def start_countdown(self, seconds=0):
		'''
		Creates and starts the countdown timer on the device.

		Args:
			seconds (int): The amount of seconds to countdown from.

		Returns:
			None
		'''
		minutes = int(seconds / 60)
		seconds = seconds % 60
		self.__request('Tools/SetTimer', {
			'Minute': minutes,
			'Second': seconds,
			'Status': 1
		})
		self.__start_countdown_time = time()

	def stop_countdown(self):
		'''
		Stops the countdown timer on the device.

		Args:
			None

		Returns:
			int: Elapsed time in seconds
		'''
		self.__request('Tools/SetTimer', {
			'Status': 0
		})
		return int(time() - self.__start_countdown_time)
	
	def clear_remote_text(self):
		'''
		Clears the remote text on the device.
		'''
		return self.__request('Draw/ClearHttpText')

	def switch(self, on=True):
		self.__request('Channel/OnOffScreen', {
			'OnOff': 1 if on else 0
		})

	def set_brightness(self, brightness):
		self.__request('Channel/SetBrightness', {
			'Brightness': clamp(brightness, 0, 100)
		})

	def set_dial(self, items, background=None, clear=True):
		'''
		Sets the dial on the device with the given items. These are networking commands on the pixoo device that manage things like temperature, weather, time, etc.
		Most of them just auto-update, so it's useful for creating custom dials (Like watchfaces, for example).
		
		Args:
			items (list(dict)): A list of items to display on the dial. Each item should have at least a 'DisplayType' type.
			background (str): The path to an image to use as the background of the dial. Default is None.
			clear (bool): Whether to send a network clear for the current text on the device or not. Default is True.

		Returns:
			None

		Note: This method is pretty raw and depends on knowing the exact parameters to send to the device. It's recommended to use the higher-level render_template method.
		If needed additional documentation [can be found on the official API](https://doc.divoom-gz.com/web/#/12?page_id=234).
		'''
		if background is not None:
			path_ext = background.split('.')[-1]
			if path_ext == 'gif':
				self._pizoo.draw_gif(background, size='auto', fill='auto')
			else:
				self._pizoo.draw_image(background)
			self._pizoo.render()
		if clear:
			self.clear_remote_text()
		processed_items = [{'TextId': index + 1, **DIAL_DEFAULT_ITEM, **item} for index, item in enumerate(items)]
		self.__request('Draw/SendHttpItemList', {
			'ItemList': processed_items
		})

	def render(self, buffer, frame_speed):
		# TODO: This may improve render speed when sending multi-frame buffers: https://www.reddit.com/r/Divoom_Products/comments/11107e1/comment/j9z776i/?utm_source=share&utm_medium=web3x&utm_name=web3xcss&utm_term=1&utm_content=share_button
		self.__pic_id += 1
		if self.__pic_id >= self.__id_limit:
			self.__reset_pic_id()
		buffer = buffer[-self._max_frames:]
		# Because of a weird bug in the Pixoo64, we need to make sure the frame speed is not below 95 or greater than 280 (290 is the max speed)
		# frame_speed = floor(clamp(frame_speed, 95, 280))
		frame_speed = floor(clamp(frame_speed, 10, 10000))
		for i, frame in enumerate(buffer):
			self.__send_frame(frame, speed=frame_speed, frame_number=len(buffer), offset=i)

	def compile_node_root_options(self, options):
		result = []
		if options['brigthness'] is not None:
			result.append((self.set_brightness, {'brightness': options['brigthness']}))
		if options['turn_screen'] is True:
			result.append((self.switch, {}))
		if options['notification'] is True:
			result.append((self.buzzer, {'duration': 2}))
		return result
	
	def __command_atributes(self, node):
		attributes = {}
		attributes['dir'] = node.attrib.get('scroll', 'left')
		attributes['dir'] = 0 if attributes['dir'] == 'left' else 1
		try:
			attributes['font'] = int(node.attrib.get('font', '2'))
		except:
			attributes['font'] = 18 if node.attrib['font'] == 'small' else 2
		attributes['speed'] = clamp(int(node.attrib.get('speed', '100')), 1, 100)
		attributes['color'] = get_color_rgb(node.attrib.get('color', '7'))
		attributes['align'] = node.attrib.get('align', 'left')
		attributes['align'] = 1 if attributes['align'] == 'left' else 2 if attributes['align'] == 'center' else 3
		if 'update' in node.attrib:
			attributes['update_time'] = int(node.attrib['update'])
		return attributes

	def compile_node(self, node, parent, inherited_props, node_props):
		tag = node.tag
		width, height = node_props['node_size']
		abs_x, abs_y = node_props['abs_coords']
		result = None
		if tag == 'message':
			text = node.text
			attributes = self.__command_atributes(node)
			result = (DisplayType.TEXT_MESSAGE, {**attributes, 'TextString': text, 'x': abs_x, 'y': abs_y, 'TextWidth': width, 'TextHeight': height})
		elif tag == 'time':
			attributes = self.__command_atributes(node)
			time_format = node.attrib.get('format', 'HH:mm:ss')
			if time_format == 'HH:mm:ss':
				dt = DisplayType.HOUR_MIN_SEC
			elif time_format == 'HH':
				dt = DisplayType.HOUR
			elif time_format == 'HH:mm':
				dt = DisplayType.HOUR_MIN
			elif time_format == 'mm':
				dt = DisplayType.MINUTE
			elif time_format == 'ss':
				dt = DisplayType.SECOND
			result = (dt, {**attributes, 'x': abs_x, 'y': abs_y, 'TextWidth': width, 'TextHeight': height})
		elif tag == 'temperature':
			attributes = self.__command_atributes(node)
			temp_format = node.attrib.get('kind', 'actual')
			if temp_format == 'actual':
				dt = DisplayType.TEMP_DIGIT
			elif temp_format == 'max':
				dt = DisplayType.TODAY_MAX_TEMP
			elif temp_format == 'min':
				dt = DisplayType.TODAY_MIN_TEMP
			result = (dt, {**attributes, 'x': abs_x, 'y': abs_y, 'TextWidth': width, 'TextHeight': height})     
		elif tag == 'weather':
			attributes = self.__command_atributes(node)
			result = (DisplayType.WEATHER_WORD, {**attributes, 'x': abs_x, 'y': abs_y, 'TextWidth': width, 'TextHeight': height})
		elif tag == 'date':
			attributes = self.__command_atributes(node)
			date_format = node.attrib.get('format', 'DD/MM/YYYY')
			if date_format == 'DD/MM/YYYY':
				dt = DisplayType.DAY_MONTH_YEAR
			elif date_format == 'DD':
				dt = DisplayType.DAY
			elif date_format == 'MM':
				dt = DisplayType.MONTH
			elif date_format == 'MMM':
				dt = DisplayType.ENG_MONTH
			elif date_format == 'YYYY':
				dt = DisplayType.YEAR
			elif date_format == 'MM/YY':
				dt = DisplayType.MONTH_YEAR
			elif date_format == 'WW':
				dt = DisplayType.ENG_WEEK_TWO
			elif date_format == 'WWW':
				dt = DisplayType.ENG_WEEK_THREE
			elif date_format == 'WWWW':
				dt = DisplayType.ENG_WEEK_FULL
			result = (dt, {**attributes, 'x': abs_x, 'y': abs_y, 'TextWidth': width, 'TextHeight': height})
		return result
	
	def render_template_items(self, items, use_cache=True):
		# workaround for a desync error with the divoom device when sending dial items
		sleep(0.2)
		items = [{'type': item[0], **item[1], 'color': tuple_to_hex(item[1]['color'])} for item in items]
		self.set_dial(items)

class ImageRenderer(Renderer):
	def __init__(self, address, pizzoo, debug, resize_factor=5, resample_method=Image.NEAREST):
		'''
		This renderer creates a static image or gif with the frames and saves it to the disk. It can be used for debugging or demo purposes.
		'''
		super().__init__(address, pizzoo, debug)
		self._size = 64
		self._max_frames = 60
		self._resize_factor = resize_factor
		self._resample_method = resample_method

	def switch(self, on=True):
		pass
	
	def render(self, buffer, frame_speed):
		'''
		The static render creates an image (if one frame) or a gif (if multiple frames) and then shows and returns it.
		'''
		buffer = buffer[-self._max_frames:]
		if len(buffer) == 1:
			image = Image.frombytes('RGB', (self._size, self._size), bytes(buffer[0]), 'raw')
			if self._resize_factor > 1:
				image = image.resize((self._size * self._resize_factor, self._size * self._resize_factor), resample=self._resample_method)
			image.save('temp.png')
			image.show()
		else:
			# Create gif with all frames
			images = []
			for frame in buffer:
				images.append(Image.frombytes('RGB', (self._size, self._size), bytes(frame), 'raw'))
			if self._resize_factor > 1:
				images = [image.resize((self._size * self._resize_factor, self._size * self._resize_factor), resample=self._resample_method) for image in images]
			images[0].save('temp.gif', save_all=True, append_images=images[1:], loop=0, duration=frame_speed)

	def __command_atributes(self, node):
		attributes = {}
		attributes['color'] = get_color_rgb(node.attrib.get('color', '7'))
		return attributes
	
	def compile_node(self, node, parent, inherited_props, node_props):
		tag = node.tag
		abs_x, abs_y = node_props['abs_coords']
		result = None
		if tag == 'message':
			text = node.text
			attributes = self.__command_atributes(node)
			result = (self._pizzoo.draw_text, {'text': text, 'xy': (abs_x, abs_y), **attributes})
		elif tag == 'time':
			attributes = self.__command_atributes(node)
			time_format = node.attrib.get('format', 'HH:mm:ss')
			time_format = time_format.replace('HH', '%H').replace('mm', '%M').replace('ss', '%S')
			text = datetime.now().strftime(time_format)
			result = (self._pizzoo.draw_text, {'text': text, 'xy': (abs_x, abs_y), **attributes})
		elif tag == 'temperature':
			attributes = self.__command_atributes(node)
			result = (self._pizzoo.draw_text, {'text': '30°C', 'xy': (abs_x, abs_y), **attributes})     
		elif tag == 'weather':
			attributes = self.__command_atributes(node)
			result = (self._pizzoo.draw_text, {'text': 'Sunny', 'xy': (abs_x, abs_y), **attributes})
		elif tag == 'date':
			attributes = self.__command_atributes(node)
			date_format = node.attrib.get('format', 'DD/MM/YYYY')
			date_format = date_format.replace('DD', '%d').replace('MM', '%m').replace('YYYY', '%Y').replace('WWW', '%a')
			text = datetime.now().strftime(date_format)
			result = (self._pizzoo.draw_text, {'text': text, 'xy': (abs_x, abs_y), **attributes})
		return result

class WindowRenderer(Renderer):
	def __init__(self, address, pizzoo, debug):
		'''
		This renderer creates a window with a canvas to render the frames on the screen. It can be used for debugging or testing purposes.
		'''
		super().__init__(address, pizzoo, debug)
		self._size = 64
		self._max_frames = 60
		self._resize_factor = 5
		self.__resize_size = (self._size * self._resize_factor, self._size * self._resize_factor)
		self._root = tk.Tk()
		self._root.title('Pizzoo emulator')
		self._root.geometry('{0}x{1}'.format(self.__resize_size[0], self.__resize_size[1]))
		self._root.attributes('-topmost', True)
		self.__canvas = tk.Canvas(self._root, width=self._size * self._resize_factor, height=self._size * self._resize_factor, bg='black')
		self.__canvas.pack()

		image = Image.new('RGB', (self._size, self._size), color='black')

		image = self._process_image(image)
		self.__image = self.__canvas.create_image(self.__resize_size[0] / 2, self.__resize_size[0] / 2, image=image)
		
		self._root.update()

	def switch(self, on=True):
		pass

	def _process_image(self, image):
		image = image.resize(self.__resize_size, Image.NEAREST)
		return ImageTk.PhotoImage(image)

	def render(self, buffer, frame_speed):
		'''
		The static render creates an image (if one frame) or a gif (if multiple frames) and then displays it on the window.
		'''
		buffer = buffer[-self._max_frames:]
		wh = self._size * self._resize_factor
		if len(buffer) == 1:
			image = Image.frombytes('RGB', (self._size, self._size), bytes(buffer[0]), 'raw')
			image = self._process_image(image)
			self.__canvas.itemconfig(self.__image, image=image)
		else:
			# Create gif with all frames
			images = []
			for frame in buffer:
				images.append(Image.frombytes('RGB', (self._size, self._size), bytes(frame), 'raw'))
			images = [image.resize((wh, wh), resample=Image.NEAREST) for image in images]
		self._root.update()

__all__ = (Renderer, Pixoo64Renderer, ImageRenderer, WindowRenderer)


from time import time, perf_counter
from PIL import Image, ImageSequence
from math import ceil, floor
from os.path import join

def tile_to_coords(tile_size, tile_x, tile_y):
	'''
	Returns the coordinates of a tile in the map based on the tile size and the tile indexes.

	Args:
		tile_size (int): the size of the tile in pixels
		tile_x (int): the x index of the tile
		tile_y (int): the y index of the tile

	Returns:
		Coordinates (tuple): a tuple with the x and y coordinates of the tile
	'''
	return (tile_x * tile_size, tile_y * tile_size)

def coords_to_tile(tile_size, x, y, use_ceil=False):
	'''
	Returns the indexes of a tile in the map based on the tile size and the coordinates.

	Args:
		tile_size (int): the size of the tile in pixels
		x (int): the x coordinate of the tile
		y (int): the y coordinate of the tile
		use_ceil (bool): if True, the indexes will be rounded up, otherwise they will be rounded down

	Returns:
		Tile indexes (tuple): a tuple with the x and y indexes of the tile
	'''
	round_method = ceil if use_ceil else floor
	return (round_method(x / tile_size), round_method(y / tile_size))

def distance(a, b):
	'''
	Returns the distance between two points in a 2D space.

	Args:
		a (tuple): a tuple with the x and y coordinates of the first point
		b (tuple): a tuple with the x and y coordinates of the second point

	Returns:
		Distance (float): the distance between the two points
	'''
	return ((b[0] - a[0]) ** 2 + (b[1] - a[1]) ** 2) ** (1 / 2)

class Camera:
	def __init__(self, game, x, y, center_on=None):
		'''
		Creates a new camera object. A camera is a portion of the map that is visible/rendered on the screen, it can be centered on an object or moved manually.
		
		Args:
			game (PizzooGame): the game object
			x (int): the x coordinate of the camera
			y (int): the y coordinate of the camera
			center_on (Actor): the object to center the camera on
		'''
		self.g = game
		self.x = x
		self.y = y
		self.width = 64
		self.height = 64
		self.center_on = center_on
		if self.center_on is not None:
			self.center(self.center_on)
	
	def center(self, target):
		'''
		Centers the camera on a target object.

		Args:
			target (Actor): the object to center the camera on
		'''
		current_map = self.g.maps[self.g.current_map]
		current_map_width = current_map.tiles_width * current_map.tile_size
		current_map_height = current_map.tiles_height * current_map.tile_size
		self.x = target.x - int(self.width / 2)
		self.y = target.y - int(self.height / 2)
		if self.x < 0:
			self.x = 0
		elif self.x + self.width > current_map_width:
			self.x = current_map_width - self.width
		if self.y < 0:
			self.y = 0
		elif self.y + self.height > current_map_height:
			self.y = current_map_height - self.height
	
	def step(self):
		'''
		Updates the camera position based on the target object if `center_on` is selected, else does nothing by default.
		'''
		if self.center_on is None:
			return
		self.center(self.center_on)

class Map:
	def __init__(self, tile_size, tiles_width, tiles_height):
		'''
		Creates a new map object. A map is a grid of tiles that can be filled with sprites.

		Args:
			tile_size (int): the size of the tiles in pixels
			tiles_width (int): the width of the map in tiles
			tiles_height (int): the height of the map in tiles
		'''
		self.tile_size = tile_size
		self.tiles_width = tiles_width
		self.tiles_height = tiles_height
		# create tiles by value not by reference
		self.tiles = [[None for x in range(tiles_width)] for y in range(tiles_height)]
		self.map_image = None
		self.map_memo = {}
	
	def draw(self, pizzoo, camera):
		'''
		Draws the map on the screen using a cropped image of the map image and a memoization based on the camera position
		
		Args:
			pizzoo (Pizzoo): the pizzoo object
			camera (Camera): the camera object
		'''
		if self.map_image is None:
			return
		# check map memo
		result = None
		index = (camera.x, camera.y, camera.width, camera.height)
		if not index in self.map_memo:
			result = self.map_image.crop((camera.x, camera.y, camera.x + camera.width, camera.y + camera.height))
			self.map_memo[index] = result
		else:
			result = self.map_memo[index]
		pizzoo.draw_image(result, (0, 0))

	def fill(self, tile):
		'''
		Fills the entire map with a single tile.

		Args:
			tile (Sprite): the tile to fill the map with
		'''
		for y in range(self.tiles_height):
			for x in range(self.tiles_width):
				if self.tiles[y][x] is None:
					self.tiles[y][x] = []
				self.tiles[y][x].append(tile)
				self.tiles[y][x].sort(key=lambda x: x.z_index)
		self.generate_map_image()

	def generate_map_image(self):
		'''
		Generates a map image based on the tiles in the map. This static image is used to render the map on the screen and is memoized for performance.	
		'''
		self.map_memo = {}
		self.map_image = Image.new('RGB', (self.tiles_width * self.tile_size, self.tiles_height * self.tile_size))
		for y in range(self.tiles_height):
			for x in range(self.tiles_width):
				if self.tiles[y][x] is not None:
					for tile in self.tiles[y][x]:
						tile_x, tile_y = tile_to_coords(self.tile_size, x, y)
						self.map_image.paste(tile.get_frame(), (tile_x, tile_y))
		return self.map_image

	def add_tile_at(self, tile, x, y):
		'''
		Adds a tile to the map at a specific position.

		Args:
			tile (Sprite): the tile to add
			x (int): the x index of the tile
			y (int): the y index of the tile
		'''
		if self.tiles[y][x] is None:
			self.tiles[y][x] = []
		self.tiles[y][x].append(tile)
		self.tiles[y][x].sort(key=lambda x: x.z_index)
	
	def add_tile(self, tile, x, y, span=(0, 0)):
		'''
		Adds a tile to the map at a specific position with a span.

		Args:
			tile (Sprite): the tile to add
			x (int): the x index of the tile
			y (int): the y index of the tile
			span (tuple): a tuple with the width and height of the tile in tiles
		'''
		for i in range(span[0]):
			for j in range(span[1]):
				self.add_tile_at(Sprite(tile.frame_src, current_frame=tile.current_frame, base_path=tile.base_path, offset=(i * self.tile_size, j * self.tile_size, self.tile_size), z_index=1), x + i, y + j)
		self.generate_map_image()

class Path():
	def __init__(self, nodes, current=0, tile_size=1, closed=False):
		self.nodes = nodes
		self.closed = closed
		self.tile_size = tile_size
		self.current = current
		self.has_ended = False

	def next(self, x, y, coords=False):
		'''
		Returns the next node in the path, if the path is closed, it will return to the first node when it reaches the end.

		Args:
			x (int): the x coordinate of the object that is following the path
			y (int): the y coordinate of the object that is following the path
			coords (bool): if True, x and y will be treated as coordinates and result will be a tuple of coordinates, otherwise it will return a tuple of tile indexes
		
		Returns:
			Next node (tuple): a tuple with the x and y indexes of the next node in the path
		'''
		if self.has_ended:
			return None
		if coords:
			x, y = coords_to_tile(self.tile_size, x, y)
		if self.current >= len(self.nodes):
			if self.closed:
				self.current = 0
			else:
				self.has_ended = True
				return None
		if self.nodes[self.current][0] == x and self.nodes[self.current][1] == y:
			self.current += 1
			if self.current >= len(self.nodes):
				if self.closed:
					self.current = 0
				else:
					self.has_ended = True
					return None
		return self.nodes[self.current] if not coords else tile_to_coords(self.tile_size, self.nodes[self.current][0], self.nodes[self.current][1])		
	
	def prev(self, x, y, coords=False):
		'''
		Returns the previous node in the path, if the path is closed, it will return to the last node when it reaches the beginning.

		Args:
			x (int): the x coordinate of the object that is following the path
			y (int): the y coordinate of the object that is following the path
			coords (bool): if True, x and y will be treated as coordinates and result will be a tuple of coordinates, otherwise it will return a tuple of tile indexes
		
		Returns:
			Previous node (tuple): a tuple with the x and y indexes of the previous node in the path
		'''
		if self.has_ended:
			return None
		if coords:
			x, y = coords_to_tile(self.tile_size, x, y)
		if self.current < 0:
			if self.closed:
				self.current = len(self.nodes) - 1
			else:
				self.has_ended = True
				return None
		if self.nodes[self.current][0] == x and self.nodes[self.current][1] == y:
			self.current -= 1
			if self.current < 0:
				if self.closed:
					self.current = len(self.nodes) - 1
				else:
					self.has_ended = True
					return None
		return self.nodes[self.current] if not coords else tile_to_coords(self.tile_size, self.nodes[self.current][0], self.nodes[self.current][1])


class PizzooGame:
	def __init__(self, pizzoo, frame_limit=5, dev=False):
		'''
		Creates a new game object. A game is a container for maps, instances, cameras, and timers. It also handles the game loop and the rendering of the game.

		Args:
			pizzoo (Pizzoo): the pizzoo object
			frame_limit (int): the frame limit of the game
			dev (bool): if True, the game will print the FPS on the console
		'''
		from pynput import keyboard
		self.pizzoo = pizzoo
		self.listener = keyboard.Listener(on_press=self.on_press)
		self.running = False
		self.frame_limit = frame_limit
		self.step_time = 1 / self.frame_limit
		self.timers = {}
		self.instances = []
		# maps is an array of maps, an every map is a multidimensional array of tiles (like a matrix)
		self.maps = []
		self.current_map = 0
		self.camera = Camera(self, 0, 0)
		self.dev = dev

	def create_camera(self, x=0, y=0, center_on=None):
		'''
		Creates a new camera object.

		Args:
			x (int): the x coordinate of the camera
			y (int): the y coordinate of the camera
			center_on (Actor): the object to center the camera on
		'''
		self.camera = Camera(self, x, y, center_on)

	def create_map(self, tile_size, tiles_width, tiles_height):
		'''
		Creates a new map with a tile size and a width and height in tiles
		
		Args:
			tile_size (int): the size of the tiles in pixels
			tiles_width (int): the width of the map in tiles
			tiles_height (int): the height of the map in tiles

		Returns:
			Map (Map): the map object
		'''
		return Map(tile_size, tiles_width, tiles_height)

	def add_instance(self, instance):
		'''
		Adds an instance to the game, the instances are sorted by the z_index attribute.

		Args:
			instance (Actor): the instance to add
		'''
		self.instances.append(instance)
		self.instances.sort(key=lambda x: x.z_index)

	def check_collision_point(self, x, y, object_x, object_y, size):
		'''
		Checks if a point is inside a bounding box.

		Args:
			x (int): the x coordinate of the point
			y (int): the y coordinate of the point
			object_x (int): the x coordinate of the bounding box
			object_y (int): the y coordinate of the bounding box
			size (int): the size of the bounding box

		Returns:
			Collision (bool): True if the point is inside the bounding box, False otherwise
		'''
		if x >= object_x and x <= object_x + size:
			if y >= object_y and y <= object_y + size:
				return True
		return False
	
	def _check_timers(self, now):
		'''
		Checks the timers and executes the callbacks if the interval has passed.

		Args:
			now (float): the current time in seconds
		'''
		new_timers = {}
		for name, timer in self.timers.items():
			if now - timer['last'] >= timer['interval']:
				timer['last'] = now
				timer['callback']()
				if timer['repeat']:
					new_timers[name] = timer
			else:
				new_timers[name] = timer
		self.timers = new_timers

	def check_collisions(self, target, class_names=None):
		'''
		Checks the collisions of an object with other objects in the game.

		Args:
			target (Actor): the object to check the collisions with
			class_names (list, str): a list of class names or a single class name to filter the collisions
		
		Returns:
			Collision list (list): a list of objects that collided with the target
		'''
		if class_names is None:
			filtered = self.instances
		else:
			if type(class_names) != list:
				class_names = [class_names]
			filtered = [instance for instance in self.instances if instance.solid and instance.__class__.__name__ in class_names]
		collision_list = []
		for instance in filtered:
			if self.collision(target, instance):
				collision_list.append(instance)
		return collision_list

	def collision(self, source, target):
		'''
		Checks if two objects are colliding.

		Args:
			source (Actor): the first object
			target (Actor): the second object
		
		Returns:
			Collision (bool): True if the objects are colliding, False otherwise
		'''
		if source.x >= target.x and source.x <= target.x + target.bounding_box_size[0]:
			if source.y >= target.y and source.y <= target.y + target.bounding_box_size[1]:
				return True
		return False

	def draw(self):
		'''
		Draws the game on the screen, rendering the maps, instances, and the UI.
		'''
		self.pizzoo.cls()
		self.maps[self.current_map].draw(self.pizzoo, self.camera)
		visible_instances = [instance for instance in self.instances if instance.visible]
		'''Render instances relative to the camera'''
		for instance in visible_instances:
			self.pizzoo.draw_image(instance.sprite.get_frame(), (int(instance.x - self.camera.x), int(instance.y - self.camera.y)))

	def draw_ui(self):
		'''
		Draws the UI on the screen, rendering the UI elements.
		'''
		pass

	def remove_instance(self, instance):
		'''
		Removes an instance from the game.

		Args:
			instance (Actor): the instance to remove

		Returns:
			Instance (Actor): the removed instance
		'''
		if instance not in self.instances:
			return
		self.instances.remove(instance)
		self.instances.sort(key=lambda x: x.z_index)
		return instance
	
	def remove_timer(self, name):
		'''
		Removes a timer from the game.

		Args:
			name (str): the name of the timer to remove

		Returns:
			Timer (dict): the removed timer
		'''
		if name not in self.timers:
			return
		selected = self.timers[name]
		del self.timers[name]
		return selected
	
	def run_task(self):
		self.step()
		self.draw()
		self.draw_ui()
		self.pizzoo.render()
		self._check_timers(time())

	def run(self):
		'''
		Runs the game loop, updating the game state and rendering the game on the screen.
		'''
		prev_time = perf_counter()
		current_time = prev_time
		count = 0
		accum = 0
		while self.running:
			prev_time = current_time
			current_time = perf_counter()
			dt = current_time - prev_time
			self.run_task()
			if self.dev:
				accum += dt
				count += 1
				if accum >= 1.0:
					accum -= 1.0
					print(f'FPS: {count}')
					count = 0
			while perf_counter() < (current_time + self.step_time):
				pass

	def start(self):
		'''
		Starts the game loop and the keyboard listener.
		'''
		self.pizzoo.switch()
		self.listener.start()
		self.running = True
		self.run()

	def stop(self):
		'''
		Stops the game loop and the keyboard listener.
		'''
		self.running = False
		self.listener.stop()
		self.pizzoo.switch(False)

	def step(self):
		'''
		Updates the game state, moving the instances, checking the collisions, and updating the camera.
		'''
		for instance in self.instances:
			instance.step()
			instance.collisions()
		self.camera.step()

	def timer(self, name, callback, interval, repeat=False, force=False):
		'''
		Creates a new timer with a name, a callback function, an interval in seconds, and a repeat flag.

		Args:
			name (str): the name of the timer
			callback (function): the callback function to execute
			interval (float): the interval in seconds
			repeat (bool): if True, the timer will repeat
			force (bool): if True, the timer will be created even if it already exists
		'''
		if name in self.timers and not force:
			return
		self.timers[name] = {
			'interval': interval,
			'callback': callback,
			'last': time(),
			'repeat': repeat
		}

	def timer_exists(self, name):
		'''
		Checks if a timer exists.

		Args:
			name (str): the name of the timer

		Returns:
			Exists (bool): True if the timer exists, False otherwise
		'''
		return name in self.timers

	def on_press(self, key):
		'''
		Handles the key press events and forwards them to the instances.

		Args:
			key (str): the key that was pressed
		'''
		key = key.char if hasattr(key, 'char') else key
		try:
			for instance in self.instances:
				instance.on_press(key)
		except AttributeError as e:
			print(e) 
			pass

class Sprite:
	def __init__(self, frame_src, base_path=None, current_frame=0, z_index=0, offset=(0, 0, 0)):
		'''
		Creates a new sprite object. A sprite is a sequence of frames that can be rendered on the screen.

		Args:
			frame_src (str, list): Can be a list of paths to images or an iterable of images (like a gif)
			base_path (str): the base path to the image files
			current_frame (int): the current frame of the sprite
			z_index (int): the z index of the sprite
			offset (tuple): a tuple with the x, y, and size of the offset to crop the image
		'''
		self.base_path = base_path
		self.frame_src = frame_src
		self.frames = self._create_frames(frame_src, offset, base_path)
		self.current_frame = current_frame
		self.z_index = z_index

	def _create_frames(self, frame_src, offset, base_path):
		'''
		Creates the frames of the sprite based on the frame source, offset, and base path.

		Args:
			frame_src (str, list): Can be a list of paths to images or an iterable of images (like a gif)
			offset (tuple): a tuple with the x, y, and size of the offset to crop the image
			base_path (str): the base path to the image files

		Returns:
			Frames (list): a list of frames
		'''
		frames = []
		if isinstance(frame_src, str):
			if base_path is not None:
				frame_src = join(base_path, frame_src)
			image = Image.open(frame_src)
			if offset != (0, 0, 0):
				image = image.crop((offset[0], offset[1], offset[0] + offset[2], offset[1] + offset[2]))
			for frame in ImageSequence.Iterator(image):
				frames.append(frame)
		else:
			for frame in frame_src:
				if base_path is not None:
					frame = join(base_path, frame)
				image = Image.open(frame)
				if offset != (0, 0, 0):
					image = image.crop((offset[0], offset[1], offset[0] + offset[2], offset[1] + offset[2]))
				frames.append(image)
		return frames
	
	def size(self):
		'''
		Returns the size of the current frame.

		Returns:
			Size (tuple): a tuple with the width and height of the current frame
		'''
		return self.frames[self.current_frame].size
			
	def next_frame(self):
		'''
		Sets the next frame as the current frame.
		'''
		if self.current_frame + 1 >= len(self.frames):
			self.current_frame = 0
		else:
			self.current_frame += 1

	def get_frame(self, index=None):
		'''
		Returns a frame based on an index.

		Args:
			index (int): the index of the frame to return

		Returns:
			Frame (PIL.Image): the frame
		'''
		if index is not None:
			return self.frames[index]
		return self.frames[self.current_frame]
	
	def set_frame(self, index):
		'''
		Sets a frame as the current frame.

		Args:
			index (int): the index of the frame to set as the current frame
		'''
		self.current_frame = index
	
	def __getitem__(self, key):
		return self.frames[key]
	
	def __len__(self):
		return len(self.frames)

class Actor:
	def __init__(self, x, y, frame_src, game, base_path=None, z_index=0, bounding_box_size=None, visible=True, solid=True):
		'''
		Creates a new actor object. An actor is an object that can be rendered on the screen and can interact with other objects.

		Args:
			x (int): the x coordinate of the actor
			y (int): the y coordinate of the actor
			frame_src (str, list): Can be a list of paths to images or an iterable of images (like a gif)
			game (PizzooGame): the game object
			base_path (str): the base path to the image files
			z_index (int): the z index of the actor
			bounding_box_size (tuple): a tuple with the width and height of the bounding box
			visible (bool): if True, the actor will be rendered on the screen
			solid (bool): if True, the actor will collide with other objects
		'''
		self.x = x
		self.y = y
		self.sprite = Sprite(frame_src, base_path=base_path)
		self.z_index = z_index
		self.visible = visible
		self.solid = solid
		self.g = game
		self.bounding_box_size = bounding_box_size if bounding_box_size is not None else self.sprite.size()

	def check_collisions(self, class_names=None):
		'''
		Checks the collisions of the actor with other objects in the game.

		Args:
			class_names (list, str): a list of class names or a single class name to filter the collisions

		Returns:
			Collision list (list): a list of objects that collided with the actor
		'''
		return self.g.check_collisions(self, class_names)
	
	def collisions(self):
		'''
		Checks the collisions of the actor with other objects in the game.
		'''
		pass
	
	def on_press(self, key):
		'''
		Handles the key press events.

		Args:
			key (str): the key that was pressed
		'''
		pass
	
	def step(self):
		'''
		Updates the actor state, moving the actor and checking the collisions.
		'''
		pass

	def destroy(self):
		'''
		Removes the actor from the game.
		'''
		self.g.remove_instance(self)

__all__ = (PizzooGame, Sprite, Actor, Path, Camera, Map, tile_to_coords, coords_to_tile, distance)


from ._constants import PICO_PALETTE

def clamp(n, minn, maxn):
	return max(min(maxn, n), minn)

def get_color_rgb(color):
	if isinstance(color, tuple):
		return color
	if isinstance(color, str):
		if color.isdigit():
			return PICO_PALETTE[int(color)]
		if color[0] == '#' and len(color) == 7:
			return tuple(int(color[i:i+2], 16) for i in (1, 3, 5))
		if color[0] == '(' and color[-1] == ')':
			return tuple(int(x) for x in color[1:-1].split(','))
	if isinstance(color, int):
		return PICO_PALETTE[color]
	raise ValueError('Invalid color format')

def tuple_to_hex(color_tuple):
	return '#%02x%02x%02x' % color_tuple

__all__ = (clamp, get_color_rgb, tuple_to_hex)


from math import floor
from PIL import Image, ImageOps
from math import floor
from xml.etree.ElementTree import ElementTree, fromstring
from ._utils import clamp, get_color_rgb
from ._renderers import Pixoo64Renderer, Renderer, ImageRenderer, WindowRenderer
from os.path import dirname, realpath, join

class Pizzoo:
	__buffer = []
	__current_frame = -1
	__max_frames = 60
	__debug = False
	__fonts = {}
	__current_dir = dirname(realpath(__file__))

	def __init__(self, address, renderer=Pixoo64Renderer, renderer_params={}, debug=False):
		'''Initialize the Pizzoo object with the given renderer. Additional parameters can be passed to the renderer.

		Args:
			address (str): The IP address of the Pixoo device, or any renderer that needs an address.
			renderer (Renderer): The renderer to use. Default is Pixoo64Renderer.
			renderer_params (dict): Additional parameters to pass to the renderer.
			debug (bool): Whether to enable debug mode or not. Default is False.

		Returns:
			None
		'''
		self.renderer = renderer(address=address, pizzoo=self, debug=debug, **renderer_params)
		self.__compute_device_specs()
		self.__debug = debug
		# Initialize buffer
		self.add_frame()
		# get current dir of this file:
		self.load_font('default', join(self.__current_dir, 'default.bdf'), False)

	def __compute_device_specs(self):
		self.size = self.renderer.get_size()
		self.pixel_count = self.size * self.size
		self.__max_frames = self.renderer.get_max_frames()
		
	def load_font(self, font_name, path, soft=True):
		'''Loads a new font on bdf format to be used on the draw_text method.

		Args:
			font_name (str): The name to identify the font.
			path (str): The path to the font file.
			soft (bool): If True, the font will be loaded when used. If False, the font will be loaded now.
		
		Returns:
			None
		'''
		if soft:
			self.__fonts[font_name] = path
		else:
			from bdfparser import Font
			self.__fonts[font_name] = Font(path)

	def load_fonts(self, fonts):
		'''Loads multiple fonts at once.

		Args:
			fonts (dict): A dictionary with the font name as key and the font path as value.

		Returns:
			None
		'''
		for font_name, path in fonts.items():
			self.load_font(font_name, path)
		
	def cls(self, rgb=(0, 0, 0)):
		'''Clears the current frame with the given color.

		Args:
			rgb (tuple(int, int, int) | int | string): The color to clear the frame with. Default is black.
		
		Returns:
			None
		'''
		rgb = get_color_rgb(rgb)
		self.__buffer[self.__current_frame] = [rgb[0], rgb[1], rgb[2]] * self.pixel_count

	def add_frame(self, rgb=(0, 0, 0)):
		'''Adds a new frame to the animation buffer.

		Args:
			rgb (tuple(int, int, int) | int | string): The color to fill the frame with. Default is black.
		
		Returns:
			None
		'''
		assert self.__current_frame <= self.__max_frames, f'Frame limit reached, push before reaching {self.__max_frames} frames'
		frame = []
		for _ in range(self.pixel_count):
			frame.extend(rgb)
		self.__buffer.append(frame)
		self.__current_frame = len(self.__buffer) - 1

	def reset_buffer(self):
		'''Resets the animation buffer, removing all frames and adding a new one.

		Returns:
			And int with the number of items removed from the buffer.
		'''
		removed_items = len(self.__buffer)
		self.__buffer = []
		self.__current_frame = -1
		self.add_frame()
		return removed_items
	
	def get_current_frame(self):
		'''Returns the current animation frame.

		Returns:
			The current animation buffer.
		'''
		return self.__buffer[self.__current_frame]
	
	def set_current_frame(self, frame):
		'''Sets the current animation frame.

		Args:
			frame (list(int)): The frame to set as the current frame.

		Returns:
			None
		'''
		self.__buffer[self.__current_frame] = frame

	def draw_pixel(self, xy, color):
		'''Draws a single pixel on the current frame at the given coordinates.

		Args:
			xy (tuple(int, int)): The coordinates to draw the pixel at.
			color (tuple(int, int, int) | int | string): The color to draw the pixel with.

		Raises:
			ValueError: If the given coordinates are out of bounds.

		Returns:
			None
		'''
		index = xy[0] + (xy[1] * self.size)
		rgb = get_color_rgb(color)
		if index < 0 or index >= self.pixel_count:
			raise ValueError(f'Invalid index given: {index} (maximum index is {self.pixel_count - 1})')
		# FIXME: Could we maybe move to rgba and use the alpha channel to determine if the pixel is on or off? -> index = index * 4
		index = index * 3
		self.__buffer[self.__current_frame][index + 1] = rgb[1]
		self.__buffer[self.__current_frame][index] = rgb[0]
		self.__buffer[self.__current_frame][index + 2] = rgb[2]

	def draw_rectangle(self, xy, width, height, color, filled=True):
		'''Draws a rectangle on the current frame at the given coordinates.

		Args:
			xy (tuple(int, int)): The coordinates of the top-left corner of the rectangle.
			width (int): The width of the rectangle.
			height (int): The height of the rectangle.
			color (tuple(int, int, int) | int | string): The color to draw the rectangle with.
			filled (bool): Whether to fill the rectangle or not.

		Returns:
			None
		'''
		for x in range(xy[0], xy[0] + width):
			for y in range(xy[1], xy[1] + height):
				if filled or x == xy[0] or x == xy[0] + width - 1 or y == xy[1] or y == xy[1] + height - 1:
					self.draw_pixel((x, y), color)

	def draw_circle(self, xy, radius, color, filled=True):
		'''Draws a circle on the current frame at the given coordinates.

		Args:
			xy (tuple(int, int)): The coordinates of the center of the circle.
			radius (int): The radius of the circle.
			color (tuple(int, int, int) | int | string): The color to draw the circle with.
			filled (bool): Whether to fill the circle or not.

		Returns:
			None
		'''
		for x in range(xy[0] - radius, xy[0] + radius + 1):
			for y in range(xy[1] - radius, xy[1] + radius + 1):
				if (x - xy[0]) ** 2 + (y - xy[1]) ** 2 <= radius ** 2:
					self.draw_pixel((x, y), color)
	
	def draw_image(self, image_or_path, xy=(0, 0), size='auto', resample_method=Image.NEAREST):
		'''Draws an image on the current frame at the given coordinates.

		Args:
			image_or_path (str | Image): The path to the image file or the Image object to draw.
			xy (tuple(int, int)): The coordinates of the top-left corner of the image.
			size (tuple(int, int) | str): The size to resize the image to. If 'auto' is given, the image will be resized to fit the screen if needed.
			resample_method (Resampling): The resample mode to use when resizing the image to fit the screen. Default is Image.NEAREST.
		
		Returns:
			None
		'''
		if isinstance(image_or_path, str):
			image = Image.open(image_or_path)
		else:
			image = image_or_path
		width, height = self.__compute_image_resize(image, size)
		image = ImageOps.fit(image, (width, height), method=resample_method, centering=(0.5, 0.5))
		image = image.convert('RGBA')
		for x in range(image.width):
			for y in range(image.height):
				rgba = image.getpixel((x, y))
				if rgba[3] > 0:
					placed_x, placed_y = x + xy[0], y + xy[1]
					if self.size - 1 < placed_x or placed_x < 0 or self.size - 1 < placed_y or placed_y < 0:
						continue
					self.draw_pixel((placed_x, placed_y), rgba)

	def draw_gif(self, gif_path, xy=(0, 0), size='auto', loop=False, resample_method=Image.NEAREST, fill='auto'):
		'''Draws a gif on the animation buffer, starting on current frame. If the gif is larger than the screen, it will be resized to fit the screen.
		
		Args:
			gif_path (str): The path to the gif file.
			xy (tuple(int, int)): The coordinates to start drawing the gif at.
			size (tuple(int, int) | str): The size to resize the gif to. If 'auto' is given, width and height of the gif will be used and resized if needed to fit the screen.
			loop (bool): Whether to loop the gif or not.
			resample_method (Resampling): The resample mode to use when resizing the gif to fit the screen. Default is Image.NEAREST.
			fill (tuple(int, int, int) | str): The color to fill the screen with before drawing the gif. If 'auto' is given, the color will be black.

		Returns:
			None
		'''
		current_frame = self.__current_frame if self.__current_frame >= 0 else 0
		gif = Image.open(gif_path)
		remaining_frames = self.__max_frames - current_frame
		total_frames = remaining_frames if loop else min(remaining_frames, gif.n_frames)
		current_gif_frame = 0
		for frame in range(0, total_frames):
			if frame >= self.__max_frames:
				break
			try:
				current_gif_frame += 1
				gif.seek(current_gif_frame)
			except EOFError:
				current_gif_frame = 0
				gif.seek(current_gif_frame)
			if fill == 'auto':
				self.cls()
			elif fill is not None:
				self.cls(fill)
			self.draw_image(gif, xy, size, resample_method)
			if frame < total_frames - 1:
				if len(self.__buffer) < self.__current_frame + 2:
					self.add_frame()
				else:
					self.__current_frame += 1
		self.__current_frame = current_frame

	def __get_font(self, font_name):
		if font_name not in self.__fonts:
			raise ValueError(f'Font "{font_name}" not found')
		if type(self.__fonts[font_name]) == str:
			from bdfparser import Font
			self.__fonts[font_name] = Font(self.__fonts[font_name])
		return self.__fonts[font_name]

	def draw_text(self, text, xy=(0, 0), font='default', color='#FFFFFF', align=0, line_width='auto', shadow=None, shadow_rgb=(0, 0, 0)):
		'''Draws a text on the current frame at the given coordinates.

		Args:
			text (str): The text to draw.
			xy (tuple(int | ('center', 'right', 'left), int | ('top', 'center', 'bottom'))): The coordinates of the top-left corner of the text.
			font (str): The name of the font to use. Default is 'default'.
			color (tuple(int, int, int) | int | string): The color to draw the text with.
			align (int): The alignment of the text. 0 is left, 1 is center and 2 is right.
			shadow (str | tuple | None): The type of shadow to add to the text. Values are 'horizontal', 'vertical', 'diagonal', tuple with displacements, or None.
        	shadow_rgb (tuple(int, int, int) | int | string): The color of the shadow.
			line_width (int): The maximum width of the text. Default is 'auto'.

		Returns:
			None
		'''
		font = self.__get_font(font)
		line_width = self.size if line_width == 'auto' else line_width
		rgb = get_color_rgb(color)
		bitmap = font.draw(text.strip(), missing='?', linelimit=line_width)
		if shadow is not None:
			shadow_rgb = get_color_rgb(shadow_rgb)
			shadow_displacement = (0, 0)
			if type(shadow) == tuple:
				shadow_displacement = shadow
			elif shadow == 'horizontal':
				shadow_displacement = (1, 0)
			elif shadow == 'vertical':
				shadow_displacement = (0, -1)
			elif shadow == 'diagonal':
				shadow_displacement = (1, -1)
			bitmap.shadow(shadow_displacement[0], shadow_displacement[1])
		text_data = bitmap.todata(2)
		width, height = bitmap.width(), bitmap.height()
		xy = self.__compute_text_coords(text, xy, width, height, (0, 0))
		for x in range(width):
			for y in range(height):
				if text_data[y][x]:
					placed_x, placed_y = x + xy[0], y + xy[1]
					if self.size - 1 < placed_x or placed_x < 0 or self.size - 1 < placed_y or placed_y < 0:
						continue
					self.draw_pixel((placed_x, placed_y), rgb if text_data[y][x] == 1 else shadow_rgb)

	def __compute_image_resize(self, image, size):
		if size == 'auto':
			width = image.width
			height = image.height
			if width > self.size:
				height = floor(height * self.size / width)
				width = self.size
			elif height > self.size:
				width = floor(width * self.size / height)
				height = self.size
		elif size == 'fill-width':
			width = self.size
			height = floor(image.height * self.size / image.width)
		elif size == 'fill-height':
			width = floor(image.width * self.size / image.height)
			height = self.size
		elif isinstance(size, tuple):
			width, height = size
		else:
			raise ValueError('Invalid size value')
		return width, height

	def draw_line(self, start, end, color):
		'''Draws a line on the current frame from the start to the end coordinates.

		Args:
			start (tuple(int, int)): The coordinates of the start of the line.
			end (tuple(int, int)): The coordinates of the end of the line.
			color (tuple(int, int, int) | int | string): The color to draw the line with.

		Returns:
			None
		'''
		x0, y0 = start
		x1, y1 = end
		dx = abs(x1 - x0)
		dy = abs(y1 - y0)
		sx = 1 if x0 < x1 else -1
		sy = 1 if y0 < y1 else -1
		err = dx - dy
		while x0 != x1 or y0 != y1:
			self.draw_pixel((x0, y0), color)
			e2 = 2 * err
			if e2 > -dy:
				err -= dy
				x0 += sx
			if e2 < dx:
				err += dx
				y0 += sy

	def render(self, frame_speed=150):
		'''Renders the current animation buffer to the Pixoo device. After that it resets the buffer.

		Take into account that only a max of 60 frames can be rendered at once. So any buffer with more than 60 frames will be truncated.

		Args:
			frame_speed (int): The speed in milliseconds per frame. Default is 150. (Only useful if more than 1 frame is being rendered)
		
		Returns:
			None
		'''
		self.renderer.render(self.__buffer, frame_speed)
		self.reset_buffer()

	def switch(self, on=True):
		'''Turns the device on or off.
		'''
		self.renderer.switch(on)
	
	def get_settings(self):
		'''Get the current settings from the device.

		Returns:
			A dict with the current settings of the device.
		'''
		return self.renderer.get_settings()
	
	def set_brightness(self, brightness):
		'''Sets the brightness of the device.

		Args:
			brightness (int): The brightness value to set. Must be between 0 and 100.

		Returns:
			None
		'''
		self.renderer.set_brightness(brightness)

	def __compute_x_text_position(self, text_width, position, padding):
		if isinstance(position, int):
			return position
		res = padding
		if text_width >= self.size:
			return res
		if position == 'left':
			res = padding
		elif position == 'center':
			res = floor((self.size - text_width) / 2)
		elif position == 'right':
			res = self.size - text_width - padding
		return res

	def __compute_y_text_position(self, text_height, position, padding):
		if isinstance(position, int):
			return position
		res = padding
		if position == 'top':
			res = padding
		elif position == 'center':
			res = floor((self.size - text_height) / 2)
		elif position == 'bottom':
			res = self.size - text_height - padding
		return res

	def __compute_text_coords(self, text, xy, width, height, paddings):
		if isinstance(xy, str) and xy == 'center':
			xy = ('center', 'center')
		return (
			self.__compute_x_text_position(width, xy[0], paddings[0]),
			self.__compute_y_text_position(height, xy[1], paddings[1])
		)
	
	def __node_position(self, x, y, position, prev_x, prev_y, rel_x, rel_y):
		'''
			- static: x and y are relative to last non-absolute parent
			- relative: as static, but absolute children are relative to this element
			- absolute: x and y are absolute to the last relative element or the screen
		'''
		abs_x = x
		abs_y = y
		position_props = {}  
		if position == 'absolute':
			abs_x = abs_x + rel_x
			abs_y = abs_y + rel_y
		else:
			abs_x = prev_x + abs_x
			abs_y = prev_y + abs_y
			position_props['x'] = abs_x
			position_props['y'] = abs_y
			if position == 'relative': 
				position_props['rel_x'] = abs_x
				position_props['rel_y'] = abs_y
		return abs_x, abs_y, position_props
	
	def __node_size(self, node, x, y, abs_x, abs_y, ancestor_props):      
		container_width = ancestor_props.get('width', self.size)
		container_height = ancestor_props.get('height', self.size)
		width = node.attrib.get('width', '100%')
		height = node.attrib.get('height', '100%')
		width = int(width[:-1]) * container_width // 100 if width[-1] == '%' else int(width)
		height = int(height[:-1]) * container_height // 100 if height[-1] == '%' else int(height)
		width = clamp(width, 1, container_width - x)
		height = clamp(height, 1, container_height - y)
		return width, height
	
	def __node_coords(self, node, ancestor_props):
		x = node.attrib.get('x', '0')
		y = node.attrib.get('y', '0')
		x = int(x[:-1]) * ancestor_props.get('width', self.size) // 100 if x[-1] == '%' else int(x)
		y = int(y[:-1]) * ancestor_props.get('height', self.size) // 100 if y[-1] == '%' else int(y)
		return x, y

	def __compile_node(self, node, parent_node=None, inherited_props=None):
		tag = node.tag
		new_props = inherited_props.copy() if inherited_props is not None else {}
		result = None
		abs_x, abs_y = 0, 0
		x, y = self.__node_coords(node, new_props)
		position = node.attrib.get('position', 'static')
		position_props = {}
		# Now this computation is used for every node
		abs_x, abs_y, position_props = self.__node_position(
			x,
			y,
			position,
			new_props.get('x', 0),
			new_props.get('y', 0),
			new_props.get('rel_x', 0),
			new_props.get('rel_y', 0)
		)
		if tag == 'section':
			width, height = self.__node_size(node, x, y, abs_x, abs_y, new_props)
			new_props['width'] = width
			new_props['height'] = height
			result = None
		elif tag == 'line':
			x2 = int(node.attrib.get('x2', '0'))
			y2 = int(node.attrib.get('y2', '0'))
			abs_x2, abs_y2, _ = self.__node_position(
				x2,
				y2,
				position,
				new_props.get('x', 0),
				new_props.get('y', 0),
				new_props.get('rel_x', 0),
				new_props.get('rel_y', 0)
			)
			color = get_color_rgb(node.attrib.get('color', '7'))
			result = (self.draw_line, {'start': (abs_x, abs_y), 'end': (abs_x2, abs_y2), 'color': color})
		elif tag == 'rectangle':
			# end position compute
			color = get_color_rgb(node.attrib.get('color', '7'))
			width, height = self.__node_size(node, x, y, abs_x, abs_y, new_props)
			new_props['width'] = width
			new_props['height'] = height
			result = (self.draw_rectangle, {'xy': (abs_x, abs_y), 'width': width, 'height': height, 'color': color, 'filled': True})
		elif tag == 'circle':
			color = get_color_rgb(node.attrib.get('color', '7'))
			radius = int(node.attrib.get('radius', '8'))
			result = (self.draw_circle, {'xy': (abs_x, abs_y), 'radius': radius, 'color': color})
		elif tag == 'text':
			text = node.text
			if text is not None:
				color = get_color_rgb(node.attrib.get('color', '7'))
				shadow = node.attrib.get('shadow', None)
				if shadow and shadow[0] == '(' and shadow[-1] == ')':
					shadow = tuple(int(x) for x in shadow[1:-1].split(','))
				shadow_color = get_color_rgb(node.attrib.get('shadowColor', '0'))
				font = node.attrib.get('font', 'default')
				wrap = node.attrib.get('wrap', 'false').lower() == 'true'
				line_width = 'auto'
				if wrap:
					width, height = self.__node_size(node, x, y, abs_x, abs_y, new_props)
					line_width = width	
				result = (self.draw_text, {'text': text, 'xy': (abs_x, abs_y), 'color': color, 'shadow': shadow, 'shadow_rgb': shadow_color, 'font': font, 'line_width': line_width})
		elif tag == 'pixel':
			color = get_color_rgb(node.attrib.get('color', '7'))
			result = (self.draw_pixel, {'xy': (abs_x, abs_y), 'color': color})
		elif tag == 'image':
			path = node.attrib.get('src', None)
			# add size
			result = (self.draw_image, {'image_or_path': path, 'xy': (abs_x, abs_y)})
		else:
			width, height = self.__node_size(node, x, y, abs_x, abs_y, new_props)
			node_props = {
				'position': position,
				'coords': (x, y),
				'abs_coords': (abs_x, abs_y),
				'rel_coords': (position_props['rel_x'], position_props['rel_y']) if 'rel_x' in position_props else None,
				'node_size': (width, height)
			}
			result = self.renderer.compile_node(node, parent_node, inherited_props, node_props)
		new_props = {**new_props, **position_props}
		return result, new_props
	
	def __compile_root_options(self, root):
		result = []
		attribs = root.attrib
		brigthness = attribs.get('brightness', None)
		turn_screen = attribs.get('turnOn', 'false').lower() == 'true'
		notification = attribs.get('notification', 'false').lower() == 'true'
		background = attribs.get('background', None)
		if background is not None:
			try:
				color = get_color_rgb(background)
				result.append((self.cls, {'rgb': color}))
			except ValueError:
				if background.endswith('.gif'):
					self.draw_gif(background, xy=(0, 0), size=(64, 64), loop=True)
				else:
					result.append((self.draw_image, {'image_or_path': background, 'xy': (0, 0), 'size': 'fill-width'}))
		else:
			result.append((self.cls, {}))
		options = {
			'brigthness': clamp(int(brigthness), 0, 100) if brigthness is not None else None,
			'notification': notification,
			'turn_screen': turn_screen,
			'background': background
		}
		result = result + self.renderer.compile_node_root_options(options)
		return result
	
	def __compile_template(self, template):
		tree = ElementTree(fromstring(template))
		root = tree.getroot()
		assert root.tag == 'pizzoo', 'Root tag must be "pizzoo"'
		commands = self.__compile_root_options(root)
		pending = [{'node': tree.getroot(), 'parent': None, 'props': {}}]
		while len(pending) > 0:
			current = pending.pop(0)
			command, props = self.__compile_node(current['node'], current['parent'], current['props'] )
			if command is not None:
				commands.append(command)
			for child in current['node']:
				pending.append({'node': child, 'parent': current, 'props': props})
		return commands
	
	def execute_commands(self, commands, renderer_items):
		for command in commands:
			if callable(command[0]):
				command[0](**command[1])
			elif renderer_items is not None:
				renderer_items.append(command)

	def render_template(self, template, use_cache=False):
		'''Renders an XML template to the given renderer.
		
		This template can have a number of valid tags that are directly supported by the library, but any other renderer can add his own nodes.
		The template must have a root tag named 'pizzoo' and can have the following tags:

		* section - A container for other elements. It can have x, y, width, height and position attributes.
		* rectangle - Draws a rectangle. It can have x, y, width, height, color and filled attributes.
		* circle - Draws a circle. It can have x, y, radius and color attributes.
		* text - Draws a text. It can have x, y, color, wrap, shadow, shadowColor and font attributes.
		* pixel - Draws a single pixel. It can have x, y and color attributes.
		* image - Draws an image. It can have x, y and src attributes.
		* line - Draws a line. It can have x, y, x2, y2 and color attributes.
		Aside from that, any container as section or rectangle can have a position attribute with the following values:
		* static - x and y are relative to last non-absolute parent
		* relative - as static, but absolute children are relative to this element
		* absolute - x and y are absolute to the last relative element or the screen

		Args:
			template (str): The XML template to render.
			use_cache (bool): Whether to use the cache or not. Default is False. (Currently not in use)

		Returns:
			None
		'''
		self.reset_buffer()
		commands = self.__compile_template(template)
		renderer_items = []
		if len(self.__buffer) == 0:
			self.execute_commands(commands, renderer_items)
		else:
			self.__current_frame = 0
			while self.__current_frame < len(self.__buffer):
				self.execute_commands(commands, renderer_items if self.__current_frame == 0 else None)
				self.__current_frame += 1
		self.render()
		if len(renderer_items) > 0:
			self.renderer.render_template_items(renderer_items, use_cache)
	
	def __getattr__(self, name):
		if hasattr(self.renderer, name) and callable(getattr(self.renderer, name)):
			return getattr(self.renderer, name)
		raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


__all__ = (Pizzoo, Renderer, Pixoo64Renderer, ImageRenderer, WindowRenderer)


from setuptools import setup
from os import path

current_folder = path.abspath(path.dirname(__file__))
long_description = open(path.join(current_folder, 'README.MD'), encoding='utf-8').read()

# https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/
setup(
    name="pizzoo",
    version="0.9.14",
    author="Pablo Huet",
    description="Pizzoo is a easy-to-use library for rendering on pixel matrix screens like the Pixoo64, featuring easy new device integration, animation tools, and XML template rendering support.",
    long_description=long_description,
	long_description_content_type='text/markdown',
	license="MIT AND (Apache-2.0 OR BSD-2-Clause)",
    keywords="pixoo, pixoo64, divoom, screen, pixel, matrix, render, buffer, LED matrix, LED, raspberry, raspberry pi",
    url="https://github.com/pabletos/pizzoo#readme",
	include_package_data=True,
    packages=['pizzoo'],
	package_dir={'pizzoo': 'pizzoo'},
	package_data={'pizzoo': ['*.bdf']},
    install_requires=[
        'requests >= 2.31.0',
        'Pillow >= 11.1.0',
		'bdfparser >= 2.2.0'
    ],
	classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Multimedia :: Graphics"
    ]
)
